"""The |pydwf.core.api.digital_io| module provides a single class: |DigitalIO|."""

from typing import Tuple

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS


class DigitalIO(AbstractDwfDeviceSubAPI):
    """The |DigitalIO| class provides access to the static digital I/O functionality of a |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently assigned to its
        public |digitalIO:link| attribute for access by the user.

    The class implements 3 generic methods: :py:meth:`reset`, :py:meth:`configure`, and :py:meth:`status`,
    and 8 methods that come in both 32- and 64-bits variants, where 32 and 64 refer to the maximum number
    of digital pins that the methods can handle.
    """

    # pylint: disable=too-many-public-methods

    def reset(self) -> None:
        """Reset all |DigitalIO| settings to default values.

        It sets the digital pins to tri-state (high impedance, not enabled) and output value to zero.

        If autoconfiguration is enabled, the values are immediately applied.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalIOReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def configure(self) -> None:
        """Configure the |DigitalIO| functionality.

        This method transfers the settings to the Digilent Waveforms device. It is not needed
        if autoconfiguration is enabled.

        Raises:
            DwfLibraryError: An error occurred while executing the *configure* operation.
        """
        result = self.lib.FDwfDigitalIOConfigure(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def status(self) -> None:
        """Read the |DigitalIO| status and input values from the device to the PC.

        The status inquiry methods that follow will return the information that was read from the device
        when this method was last called.

        Note that the |DigitalIO| functionality is not managed by a state machine,
        so this method does not return a value.

        Raises:
            DwfLibraryError: An error occurred while executing the *status* operation.
        """
        result = self.lib.FDwfDigitalIOStatus(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputEnableInfo(self) -> int:
        """Get the digital pins that can be enabled for output as a bitmask.

        The *output enable state* of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputEnableInfo64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_enable_mask = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOOutputEnableInfo(self.hdwf, c_output_enable_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_enable_mask = c_output_enable_mask.value
        return output_enable_mask

    def outputEnableSet(self, output_enable: int) -> None:
        """Set the digital pins that are enabled for output as a bitmask.

        The *output enable* state of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputEnableSet64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Parameters:
            output_enable (int): A bitmask of pins that will be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIOOutputEnableSet(self.hdwf, output_enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputEnableGet(self) -> int:
        """Get the digital pins that are enabled for output as a bitmask.

        The *output enable* state of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputEnableGet64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently configured as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_enable = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOOutputEnableGet(self.hdwf, c_output_enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_enable = c_output_enable.value
        return output_enable

    def outputInfo(self) -> int:
        """Get the digital pins that can be used as outputs, i.e., driven high or low, as a bitmask.

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputInfo64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_mask = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOOutputInfo(self.hdwf, c_output_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_mask = c_output_mask.value
        return output_mask

    def outputSet(self, output: int) -> None:
        """Set the digital pins that are currently driven high as a bitmask.

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputSet64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Parameters:
            output (int): A bitmask of pins that will be driven high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIOOutputSet(self.hdwf, output)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputGet(self) -> int:
        """Get the digital pins that are currently driven high as a bitmask.

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`outputGet64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently set to high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOOutputGet(self.hdwf, c_output)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output = c_output.value
        return output

    def pullInfo(self) -> Tuple[int, int]:
        """Get pull info.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_pull_up_mask = typespec_ctypes.c_unsigned_int()
        c_pull_down_mask = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOPullInfo(self.hdwf, c_pull_up_mask, c_pull_down_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        pull_up_mask = c_pull_up_mask.value
        pull_down_mask = c_pull_down_mask.value
        return (pull_up_mask, pull_down_mask)

    def pullSet(self, pull_up_mask: int, pull_down_mask: int) -> None:
        """Set pull-up and pull-down channels.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIOPullSet(self.hdwf, pull_up_mask, pull_down_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def pullGet(self) -> Tuple[int, int]:
        """Get pull up/down configuration.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_pull_up_mask = typespec_ctypes.c_unsigned_int()
        c_pull_down_mask = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOPullGet(self.hdwf, c_pull_up_mask, c_pull_down_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        pull_up_mask = c_pull_up_mask.value
        pull_down_mask = c_pull_down_mask.value
        return (pull_up_mask, pull_down_mask)

    def driveInfo(self) -> Tuple[float, float, int, int]:
        """Get drive info.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_amp_min = typespec_ctypes.c_double()
        c_amp_max = typespec_ctypes.c_double()
        c_amp_steps = typespec_ctypes.c_int()
        c_slew_steps = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalIODriveInfo(self.hdwf, c_amp_min, c_amp_max, c_amp_steps, c_slew_steps)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        amp_min = c_amp_min.value
        amp_max = c_amp_max.value
        amp_steps = c_amp_steps.value
        slew_steps = c_slew_steps.value
        return (amp_min, amp_max, amp_steps, slew_steps)

    def driveSet(self, channel: int, amp: float, slew: int) -> None:
        """Set channel drive.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIODriveSet(self.hdwf, channel, amp, slew)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()


    def driveGet(self, channel: int) -> Tuple[float, int]:
        """Get channel drive settings.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_amp = typespec_ctypes.c_double()
        c_slew = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalIODriveInfo(self.hdwf, channel, c_amp, c_slew)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        amp = c_amp.value
        slew = c_slew.value
        return (amp, slew)

    def inputInfo(self) -> int:
        """Return the digital pins that can be used for input on the device as a bitmask.

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`inputInfo64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as inputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_input_mask = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOInputInfo(self.hdwf, c_input_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        input_mask = c_input_mask.value
        return input_mask

    def inputStatus(self) -> int:
        """Return the current state of the digital input pins on the device as a bitmask.

        Before calling this method, call the :py:meth:`status` method to read the current digital input
        status from the device.

        This is the 32-bits version of this method. For the 64-bits version, see :py:meth:`inputStatus64`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently read as high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_input = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalIOInputStatus(self.hdwf, c_input)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        input_ = c_input.value
        return input_

    def outputEnableInfo64(self) -> int:
        """Get the digital pins that can be enabled for output as a bitmask.

        The *output enable state* of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputEnableInfo`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_enable_mask = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOOutputEnableInfo64(self.hdwf, c_output_enable_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_enable_mask = c_output_enable_mask.value
        return output_enable_mask

    def outputEnableSet64(self, output_enable: int) -> None:
        """Set the digital pins that are enabled for output as a bitmask.

        The *output enable* state of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputEnableSet`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Parameters:
            output_enable (int): A bitmask of pins that will be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIOOutputEnableSet64(self.hdwf, output_enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputEnableGet64(self) -> int:
        """Get the digital pins that are enabled for output as a bitmask.

        The *output enable* state of a pin determines if it is driven as an output.
        If not, it is in high impedance (also known as high Z) mode.

        Only digital pins that are used as outputs should be enabled; digital pins that are used as inputs
        should remain disabled (the default state after reset).

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputEnableGet`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently configured as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_enable = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOOutputEnableGet64(self.hdwf, c_output_enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_enable = c_output_enable.value
        return output_enable

    def outputInfo64(self) -> int:
        """Get the digital pins that can be used as outputs, i.e., driven high or low, as a bitmask.

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputInfo`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as outputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_mask = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOOutputInfo64(self.hdwf, c_output_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_mask = c_output_mask.value
        return output_mask

    def outputSet64(self, output: int) -> None:
        """Set the digital pins that are currently driven high as a bitmask.

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputSet`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Parameters:
            output (int): A bitmask of pins that will be driven high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalIOOutputSet64(self.hdwf, output)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputGet64(self) -> int:
        """Get the digital pins that are currently driven high as a bitmask.

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`outputGet`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently set to high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOOutputGet64(self.hdwf, c_output)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output = c_output.value
        return output

    def inputInfo64(self) -> int:
        """Return the digital pins that can be used for input on the device as a bitmask.

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`inputInfo`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that can be used as inputs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_input_mask = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOInputInfo64(self.hdwf, c_input_mask)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        input_mask = c_input_mask.value
        return input_mask

    def inputStatus64(self) -> int:
        """Return the current state of the digital input pins on the device as a bitmask.

        Before calling this method, call the :py:meth:`status` method to read the current digital input
        status from the device.

        This is the 64-bits version of this method. For the 32-bits version, see :py:meth:`inputStatus`.
        The 32 and 64 bits refer here to the maximum number of digital pins that the methods can handle.

        Returns:
            int: A bitmask of pins that are currently read as high.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_input = typespec_ctypes.c_unsigned_long_long()
        result = self.lib.FDwfDigitalIOInputStatus64(self.hdwf, c_input)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        input_ = c_input.value
        return input_



"""The |pydwf.core.api.digital_out| module provides a single class: |DigitalOut|."""

from typing import Tuple, List
import ctypes

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.enum_types import (DwfTriggerSource, DwfTriggerSlope, DwfDigitalOutOutput, DwfDigitalOutType,
                                             DwfState, DwfDigitalOutIdle)
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.auxiliary.exceptions import PyDwfError

# pylint: disable=too-many-lines

class DigitalOut(AbstractDwfDeviceSubAPI):
    """The |DigitalOut| class provides access to the digital output (pattern generator) instrument of a
    |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently assigned to its
        public |digitalOut:link| attribute for access by the user.
    """

    # pylint: disable=too-many-public-methods

    ###################################################################################################################
    #                                                                                                                 #
    #                                               INSTRUMENT CONTROL                                                #
    #                                                                                                                 #
    ###################################################################################################################

    def reset(self) -> None:
        """Reset the |DigitalOut| instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalOutReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def configure(self, start: bool) -> None:
        """Start or stop the |DigitalOut| instrument.

        Parameters:
            start (int): Whether to start/stop the instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the *configure* operation.
        """
        result = self.lib.FDwfDigitalOutConfigure(self.hdwf, start)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def status(self) -> DwfState:
        """Return the status of the |DigitalOut| instrument.

        This method performs a status request to the |DigitalOut| instrument and receives its response.

        Returns:
            DwfState: The status of the |DigitalOut| instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the *status* operation.
        """
        c_status = typespec_ctypes.DwfState()
        result = self.lib.FDwfDigitalOutStatus(self.hdwf, c_status)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        status_ = DwfState(c_status.value)
        return status_

    def statusOutput(self) -> Tuple[int, int]:
        """Get output status.
        
        Notice:
            This function is not documented in the official documentation, but it is present in the header file.
        """
        c_value = typespec_ctypes.c_unsigned_int()
        c_enable = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutStatus(self.hdwf, c_value, c_enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        value = c_value.value
        enable = c_enable.value
        return (value, enable)

    ###################################################################################################################
    #                                                                                                                 #
    #                                                  CHANNEL COUNT                                                  #
    #                                                                                                                 #
    ###################################################################################################################

    def count(self) -> int:
        """Get the |DigitalOut| instrument channel (digital pin) count.

        Returns:
            int: The number of digital output channels.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_channel_count = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutCount(self.hdwf, c_channel_count)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        channel_count = c_channel_count.value
        return channel_count

    ###################################################################################################################
    #                                                                                                                 #
    #                                             STATE MACHINE SETTINGS                                              #
    #                                                                                                                 #
    ###################################################################################################################

    def waitInfo(self) -> Tuple[float, float]:
        """Get the |DigitalOut| instrument range for the |Wait:link| state duration, in seconds.

        Returns:
            Tuple[float, float]: A tuple containing the minimal and maximal configurable |Wait:link| state duration,
            in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_wait_duration_min = typespec_ctypes.c_double()
        c_wait_duration_max = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutWaitInfo(self.hdwf, c_wait_duration_min, c_wait_duration_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        wait_duration_min = c_wait_duration_min.value
        wait_duration_max = c_wait_duration_max.value
        return (wait_duration_min, wait_duration_max)

    def waitSet(self, wait_duration: float) -> None:
        """Set the |DigitalOut| instrument |Wait:link| state duration, in seconds.

        Parameters:
            wait_duration (float): Digital-out |Wait:link| state duration, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutWaitSet(self.hdwf, wait_duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def waitGet(self) -> float:
        """Get |DigitalOut| instrument |Wait:link| state duration, in seconds.

        Returns:
            float: The |Wait:link| state duration, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_wait_duration = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutWaitGet(self.hdwf, c_wait_duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        wait_duration = c_wait_duration.value
        return wait_duration

    def runInfo(self) -> Tuple[float, float]:
        """Get the |DigitalOut| instrument range for the |Running:link| state duration, in seconds.

        Returns:
            Tuple[float, float]: A tuple containing the minimal and maximal |Running:link| state duration, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_run_duration_min = typespec_ctypes.c_double()
        c_run_duration_max = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutRunInfo(self.hdwf, c_run_duration_min, c_run_duration_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        run_duration_min = c_run_duration_min.value
        run_duration_max = c_run_duration_max.value
        return (run_duration_min, run_duration_max)

    def runSet(self, run_duration: float) -> None:
        """Set the |DigitalOut| instrument |Running:link| state duration, in seconds.

        Parameters:
            run_duration: The |Running:link| state duration, in seconds.

                The value 0 is special; it means *forever*.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutRunSet(self.hdwf, run_duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def runGet(self) -> float:
        """Get the |DigitalOut| instrument |Running:link| state duration, in seconds.

        Returns:
            float: The |Running:link| state duration, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_run_duration = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutRunGet(self.hdwf, c_run_duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        run_duration = c_run_duration.value
        return run_duration

    def runStatus(self) -> int:
        """Get the |DigitalOut| instrument |Running:link| state duration time left,
        in clock cycles.

        This value is internally expressed as an integer with 48-bit resolution,
        and is measured in integer clock cycles. The C API returns it as a
        double-precision floating point number, to avoid using 64-bit integers.

        Use the :py:meth:`internalClockInfo` method to retrieve the clock frequency.

        Returns:
            int: The number of clock cycles until the nest state transition of the
            |DigitalOut| instrument's state machine.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_run_status = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutRunStatus(self.hdwf, c_run_status)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        if not c_run_status.value.is_integer():
            raise PyDwfError("Bad c_run_status value.")

        run_status = int(c_run_status.value)
        return run_status

    def repeatTriggerSet(self, repeat_trigger_flag: bool) -> None:
        """Specify if each |DigitalOut| pulse sequence run should wait for its own trigger.

        Parameters:
            repeat_trigger_flag (bool): If True, not only the first, both also every successive run of the
                pulse output sequence will wait until it receives a trigger.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutRepeatTriggerSet(self.hdwf, repeat_trigger_flag)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def repeatTriggerGet(self) -> bool:
        """Get if each |DigitalOut| pulse sequence run should wait for its own trigger.

        Returns:
            bool: If True, not only the first, both also every successive run of the pulse output sequence
            will wait until it receives a trigger.
        """
        c_repeat_trigger = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutRepeatTriggerGet(self.hdwf, c_repeat_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repeat_trigger_flag = bool(c_repeat_trigger.value)
        return repeat_trigger_flag

    def repeatInfo(self) -> Tuple[int, int]:
        """Get the |DigitalOut| minimal and maximal repeat count for pulse-sequence runs.

        Returns:
            Tuple[int, int]: A tuple containing the minimal and maximal repeat count for
            digital-out pulse-sequence runs.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_repeat_min = typespec_ctypes.c_unsigned_int()
        c_repeat_max = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutRepeatInfo(self.hdwf, c_repeat_min, c_repeat_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repeat_min = c_repeat_min.value
        repeat_max = c_repeat_max.value
        return (repeat_min, repeat_max)

    def repeatSet(self, repeat: int) -> None:
        """Set the |DigitalOut| repeat count for pulse-sequence runs.

        Parameters:
            repeat (int): Repeat count.
                The value 0 is special; it means *forever*.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutRepeatSet(self.hdwf, repeat)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def repeatGet(self) -> int:
        """Set the |DigitalOut| repeat count for pulse-sequence runs.

        Returns:
            int: Repeat count.
            The value 0 is special; it means *forever*.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_repeat = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutRepeatGet(self.hdwf, c_repeat)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repeat = c_repeat.value
        return repeat

    def repeatStatus(self) -> int:
        """Get the |DigitalOut| count of repeats remaining for the currently active output sequence.

        This number counts down as a digital output sequence is active.

        Returns:
            int: The repeat count status.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_repeat_status = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutRepeatStatus(self.hdwf, c_repeat_status)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repeat_status = c_repeat_status.value
        return repeat_status

    def repetitionInfo(self) -> int:
        """Get repetition info."""
        c_repetition_max = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutRepetitionInfo(self.hdwf, c_repetition_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repetition_max = c_repetition_max.value
        return repetition_max

    def repetitionSet(self, channel_index: int, repeat: int) -> None:
        """Set repetition count."""
        result = self.lib.FDwfDigitalOutRepetitionSet(self.hdwf, channel_index, repeat)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def repetitionGet(self, channel_index: int) -> int:
        """Get repetition count."""
        c_repeat = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutRepetitionGet(self.hdwf, channel_index, c_repeat)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        repeat = c_repeat.value
        return repeat

    ###################################################################################################################
    #                                                                                                                 #
    #                                              TRIGGER CONFIGURATION                                              #
    #                                                                                                                 #
    ###################################################################################################################

    def triggerSourceInfo(self) -> List[DwfTriggerSource]:
        """Get the valid |DigitalOut| trigger sources.

        Warning:
            **This method is obsolete.**

            Use the generic DeviceControl.triggerInfo() method instead.

        Returns:
            List[DwfTriggerSource]: A list of valid the trigger sources.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_source_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutTriggerSourceInfo(self.hdwf, c_trigger_source_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_source_bitset = c_trigger_source_bitset.value
        trigger_source_list = [trigger_source for trigger_source in DwfTriggerSource
                               if trigger_source_bitset & (1 << trigger_source.value)]
        return trigger_source_list

    def triggerSourceSet(self, trigger_source: DwfTriggerSource) -> None:
        """Set the |DigitalOut| trigger source.

        Parameters:
            trigger_source (DwfTriggerSource): The trigger source to be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutTriggerSourceSet(self.hdwf, trigger_source.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerSourceGet(self) -> DwfTriggerSource:
        """Get the currently selected instrument trigger source.

        Returns:
            DwfTriggerSource: The currently selected instrument trigger source.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_source = typespec_ctypes.DwfTriggerSource()
        result = self.lib.FDwfDigitalOutTriggerSourceGet(self.hdwf, c_trigger_source)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_source = DwfTriggerSource(c_trigger_source.value)
        return trigger_source

    def triggerSlopeSet(self, trigger_slope: DwfTriggerSlope) -> None:
        """Select the |DigitalOut| instrument trigger slope.

        Parameters:
            trigger_slope (DwfTriggerSlope): The trigger slope to be selected.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutTriggerSlopeSet(self.hdwf, trigger_slope.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerSlopeGet(self) -> DwfTriggerSlope:
        """Get the currently configured |DigitalOut| instrument trigger slope.

        Returns:
            DwfTriggerSlope: The currently selected |DigitalOut| instrument trigger slope.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_slope = typespec_ctypes.DwfTriggerSlope()
        result = self.lib.FDwfDigitalOutTriggerSlopeGet(self.hdwf, c_trigger_slope)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_slope = DwfTriggerSlope(c_trigger_slope.value)
        return trigger_slope

    ###################################################################################################################
    #                                                                                                                 #
    #                                                 OUTPUT SETTINGS                                                 #
    #                                                                                                                 #
    ###################################################################################################################

    def enableSet(self, channel_index: int, enable_flag: bool) -> None:
        """Enable or disable a |DigitalOut| channel (pin).

        Parameters:
            channel_index (int): The digital pin to enable or disable.
            enable_flag (bool): Whether to enable or disable the digital output.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutEnableSet(self.hdwf, channel_index, enable_flag)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def enableGet(self, channel_index: int) -> bool:
        """Check if a specific |DigitalOut| channel (pin) is enabled for output.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            bool: Whether the digital pin is enabled as an output.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_enable_flag = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutEnableGet(self.hdwf, channel_index, c_enable_flag)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        enable_flag = bool(c_enable_flag.value)
        return enable_flag

    def outputInfo(self, channel_index: int) -> List[DwfDigitalOutOutput]:
        """Get valid |DigitalOut| output choices (e.g. Push/Pull, tristate).

        Returns:
            List[DwfDigitalOutOutput]: A list of valid output settings.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutOutputInfo(self.hdwf, channel_index, c_output_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_bitset = c_output_bitset.value
        output_list = [output for output in DwfDigitalOutOutput if output_bitset & (1 << output.value)]
        return output_list

    def outputSet(self, channel_index: int, output_value: DwfDigitalOutOutput) -> None:
        """Set |DigitalOut| output choice (e.g. Push/Pull, tristate).

        Parameters:
            channel_index (int): The digital pin.
            output_value (DwfDigitalOutOutput): The digital output setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutOutputSet(self.hdwf, channel_index, output_value.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def outputGet(self, channel_index: int) -> DwfDigitalOutOutput:
        """Get currently configured |DigitalOut| output (e.g. Push/Pull, tristate).

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            DwfDigitalOutOutput: The digital output setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_value = typespec_ctypes.DwfDigitalOutOutput()
        result = self.lib.FDwfDigitalOutOutputGet(self.hdwf, channel_index, c_output_value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_value = DwfDigitalOutOutput(c_output_value.value)
        return output_value

    def typeInfo(self, channel_index: int) -> List[DwfDigitalOutType]:
        """Get a list of valid |DigitalOut| output channel types.

        Returns:
            List[DwfDigitalOutType]: A list of valid digital output channel types.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_type_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutTypeInfo(self.hdwf, channel_index, c_type_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        type_bitset = c_type_bitset.value
        type_list = [type_ for type_ in DwfDigitalOutType if type_bitset & (1 << type_.value)]
        return type_list

    def typeSet(self, channel_index: int, output_type: DwfDigitalOutType) -> None:
        """Select the |DigitalOut| output channel type.

        Parameters:
            channel_index (int): The digital pin.
            output_type (DwfDigitalOutType): The digital output channel type.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutTypeSet(self.hdwf, channel_index, output_type.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def typeGet(self, channel_index: int) -> DwfDigitalOutType:
        """Get the currently selected |DigitalOut| output channel type.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            DwfDigitalOutType: The digital output channel type.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_output_type = typespec_ctypes.DwfDigitalOutType()
        result = self.lib.FDwfDigitalOutTypeGet(self.hdwf, channel_index, c_output_type)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        output_type = DwfDigitalOutType(c_output_type.value)
        return output_type

    def idleInfo(self, channel_index: int) -> List[DwfDigitalOutIdle]:
        """Get valid |DigitalOut| idle output values.

        Returns:
            List[DwfDigitalOutIdle]: A list of valid idle output values.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_idle_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalOutIdleInfo(self.hdwf, channel_index, c_idle_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        idle_bitset = c_idle_bitset.value
        idle_list = [idle for idle in DwfDigitalOutIdle if idle_bitset & (1 << idle.value)]
        return idle_list

    def idleSet(self, channel_index: int, idle_mode: DwfDigitalOutIdle) -> None:
        """Set the |DigitalOut| idle output value.

        Parameters:
            channel_index (int): The digital pin.
            idle_mode (DwfDigitalOutIdle): The idle output value.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutIdleSet(self.hdwf, channel_index, idle_mode.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def idleGet(self, channel_index: int) -> DwfDigitalOutIdle:
        """Get the currently configured idle output value.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            DwfDigitalOutIdle: The currently configured idle output value.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_idle_mode = typespec_ctypes.DwfDigitalOutIdle()
        result = self.lib.FDwfDigitalOutIdleGet(self.hdwf, channel_index, c_idle_mode)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        idle_mode = DwfDigitalOutIdle(c_idle_mode.value)
        return idle_mode

    ###################################################################################################################
    #                                                                                                                 #
    #                                        OUTPUT PATTERN TIMING DEFINITION                                         #
    #                                                                                                                 #
    ###################################################################################################################

    def internalClockInfo(self) -> float:
        """Get the |DigitalOut| instrument clock frequency.

        Returns:
            float: The digital-out clock frequency, in Hz.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_internal_clock_frequency = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalOutInternalClockInfo(self.hdwf, c_internal_clock_frequency)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        internal_clock_frequency = c_internal_clock_frequency.value
        return internal_clock_frequency

    def dividerInfo(self, channel_index: int) -> Tuple[int, int]:
        """Get the |DigitalOut| divider value range.

        Returns:
            Tuple[int, int]: The range of valid divider settings.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_divider_min = typespec_ctypes.c_unsigned_int()
        c_divider_max = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutDividerInfo(
            self.hdwf,
            channel_index,
            c_divider_min,
            c_divider_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        divider_min = c_divider_min.value
        divider_max = c_divider_max.value
        return (divider_min, divider_max)

    def dividerSet(self, channel_index: int, divider: int) -> None:
        """Set the |DigitalOut| divider value.

        Parameters:
            channel_index (int): The digital pin.
            divider (int): The divider setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutDividerSet(self.hdwf, channel_index, divider)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def dividerGet(self, channel_index: int) -> int:
        """Get the currently configured |DigitalOut| divider value.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            int: The divider setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_divider = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutDividerGet(self.hdwf, channel_index, c_divider)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        divider = c_divider.value
        return divider

    def dividerInitSet(self, channel_index: int, divider_init: int) -> None:
        """Set the |DigitalOut| initial divider value.

        Parameters:
            channel_index (int): The digital pin.
            divider_init (int): The divider init setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutDividerInitSet(self.hdwf, channel_index, divider_init)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def dividerInitGet(self, channel_index: int) -> int:
        """Get the currently configured |DigitalOut| initial divider value.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            int: The divider init setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_divider_init = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutDividerInitGet(self.hdwf, channel_index, c_divider_init)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        divider_init = c_divider_init.value
        return divider_init

    def counterInfo(self, channel_index: int) -> Tuple[int, int]:
        """Get the |DigitalOut| counter value range.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            Tuple[int, int]: The range of valid counter settings.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_counter_min = typespec_ctypes.c_unsigned_int()
        c_counter_max = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutCounterInfo(
            self.hdwf,
            channel_index,
            c_counter_min,
            c_counter_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        counter_min = c_counter_min.value
        counter_max = c_counter_max.value
        return (counter_min, counter_max)

    def counterSet(self, channel_index: int, low_count: int, high_count: int) -> None:
        """Set the |DigitalOut| counter durations for both the low and high signal output levels.

        Parameters:
            channel_index (int): The digital pin.
            low_count (int): The number of cycles the signal should be Low.
            high_count (int): The number of cycles the signal should be High.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutCounterSet(self.hdwf, channel_index, low_count, high_count)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def counterGet(self, channel_index: int) -> Tuple[int, int]:
        """Get the |DigitalOut| counter durations for both the low and high signal output levels.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            Tuple[int, int]: The number of cycles the signal should be Low, High.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_low_count = typespec_ctypes.c_unsigned_int()
        c_high_count = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutCounterGet(self.hdwf, channel_index, c_low_count, c_high_count)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        low_count = c_low_count.value
        high_count = c_high_count.value
        return (low_count, high_count)

    def counterInitSet(self, channel_index: int, high: bool, counter_init: int) -> None:
        """Set the |DigitalOut| initial signal value and initial counter value.

        Parameters:
            channel_index (int): The digital pin.
            high (bool): Whether to start High (True) or Low (False).
            counter_init (int): The initial counter setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutCounterInitSet(
            self.hdwf,
            channel_index,
            high, counter_init)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def counterInitGet(self, channel_index: int) -> Tuple[bool, int]:
        """Get the |DigitalOut| initial signal value and initial counter value.

        Parameters:
            channel_index (int): The digital pin.

        Returns:
            Tuple [bool, int]: Whether to start High (True) or Low (False),
            and the initial counter setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_high = typespec_ctypes.c_int()
        c_counter_init = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutCounterInitGet(
            self.hdwf,
            channel_index, c_high,
            c_counter_init)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        high = bool(c_high.value)
        counter_init = c_counter_init.value
        return (high, counter_init)

    ###################################################################################################################
    #                                                                                                                 #
    #                                                  DATA PLAYBACK                                                  #
    #                                                                                                                 #
    ###################################################################################################################

    def dataInfo(self, channel_index: int) -> int:
        """Return the maximum buffer size for the specified |DigitalOut| channel,
        i.e., the number of custom data bits.

        Parameters:
            channel_index (int): the channel for which to obtain the data bits count.

        Returns:
            int: The number of custom data bits that can be specified for the channel.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_max_data_bits = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalOutDataInfo(self.hdwf, channel_index, c_max_data_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        max_data_bits = c_max_data_bits.value
        return max_data_bits

    def dataSet(self, channel_index: int, bits: str, tristate: bool = False) -> None:
        """Set the |DigitalOut| arbitrary output data.

        This function also sets the counter initial, low and high value, according to the
        number of bits. The data-bits are sent out in LSB-first order.

        Parameters:
            channel_index (int): the channel for which to set the output data.
            bits (str): The bits, as a string.
            tristate (bool): Whether to interpret the string as a tristate signal.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        if tristate:
            bits = bits.replace('1', '11').replace('0', '01').replace('Z', '00')

        # For tristate output, the count of bits is the total number of output values (I/O)
        # and output enable (OE) bits, which should be an even number.

        count_of_bits = len(bits)

        octets = []
        while len(bits) > 0:
            octet_str = bits[:8]
            octet = int(octet_str[::-1], 2)
            octets.append(octet)
            bits = bits[8:]

        octets_as_bytes = bytes(octets)

        result = self.lib.FDwfDigitalOutDataSet(self.hdwf, channel_index, octets_as_bytes, count_of_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def playDataSet(self, bits: str, bits_per_sample: int, count_of_samples: int) -> None:
        """Set the |DigitalOut| playback data.

        The output can be PushPull, OpenDrain, or OpenSource. Tristate data is not supported.

        Note:
            The DWF documentation explicitly states that this function is supported by the Digital Discovery.
            (So, by implication, it's probably not supported on anything else.)

        Parameters:
            bits (str): string of '0' and '1' characters. Its length should be (bits_per_sample * count_of_samples).
            bits_per_sample (int): Bits per sample, should be 1, 2, 4, 8, or 16.
            count_of_samples (int): Number of samples.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """

        if len(bits) != (bits_per_sample * count_of_samples):
            raise PyDwfError("Unexpected number of bits.")

        octets = []
        while len(bits) > 0:
            octet_str = bits[:8]
            octet = int(octet_str[::-1], 2)
            octets.append(octet)
            bits = bits[8:]

        # We convert the 'octets' input data to an array of ctypes unsigned chars.
        # This silences a spurious mypy warning.

        octet_ctypes_type = (ctypes.c_ubyte * len(octets))  # pylint: disable=superfluous-parens
        octet_ctypes = octet_ctypes_type(*octets)

        result = self.lib.FDwfDigitalOutPlayDataSet(
            self.hdwf,
            octet_ctypes,
            bits_per_sample,
            count_of_samples)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def playUpdateSet(self, bits: str, index_of_sample: int, count_of_samples: int) -> None:
        """Set the |DigitalOut| playback data.

        ToDo:
           This function is not sufficiently documented at this time.
        """

        if len(bits) % 8 != 0:
            raise PyDwfError("Unexpected number of bits.")

        octets = []
        while len(bits) > 0:
            octet_str = bits[:8]
            octet = int(octet_str[::-1], 2)
            octets.append(octet)
            bits = bits[8:]

        # We convert the 'octets' input data to an array of ctypes unsigned chars.
        # This silences a spurious mypy warning.

        octet_ctypes_type = (ctypes.c_ubyte * len(octets))  # pylint: disable=superfluous-parens
        octet_ctypes = octet_ctypes_type(*octets)

        result = self.lib.FDwfDigitalOutPlayUpdateSet(
            self.hdwf,
            octet_ctypes,
            index_of_sample,
            count_of_samples)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def playRateSet(self, playback_rate: float) -> None:
        """Set the |DigitalOut| playback rate, in Hz.

        Note:
            The DWF documentation explicitly states that this function is supported by the Digital Discovery.
            (So, by implication, it's probably not supported on anything else.)

        Parameters:
            playback_rate (float): The playback rate, in Hz.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalOutPlayRateSet(self.hdwf, playback_rate)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()



"""The |pydwf.core.api.digital_in| module provides a single class: |DigitalIn|."""

# pylint: disable=too-many-lines

from typing import Tuple, List

import numpy as np

from pydwf.core.dwf_device_subapi import AbstractDwfDeviceSubAPI

from pydwf.core.auxiliary.enum_types import (DwfDigitalInSampleMode, DwfDigitalInClockSource,
                                             DwfAcquisitionMode, DwfTriggerSlope, DwfState, DwfTriggerSource)
from pydwf.core.auxiliary.typespec_ctypes import typespec_ctypes
from pydwf.core.auxiliary.constants import RESULT_SUCCESS
from pydwf.core.auxiliary.exceptions import PyDwfError


class DigitalIn(AbstractDwfDeviceSubAPI):
    """The |DigitalIn| class provides access to the digital input (logic analyzer) instrument of a |DwfDevice:link|.

    Attention:
        Users of |pydwf| should not create instances of this class directly.

        It is instantiated during initialization of a |DwfDevice| and subsequently assigned to its public
        |digitalIn:link| attribute for access by the user.
    """

    # pylint: disable=too-many-public-methods

    ###################################################################################################################
    #                                                                                                                 #
    #                                               INSTRUMENT CONTROL                                                #
    #                                                                                                                 #
    ###################################################################################################################

    def reset(self) -> None:
        """Reset the |DigitalIn| instrument parameters to default values.

        If autoconfiguration is enabled, the reset values will be used immediately.

        Raises:
            DwfLibraryError: An error occurred while executing the *reset* operation.
        """
        result = self.lib.FDwfDigitalInReset(self.hdwf)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def configure(self, reconfigure: bool, start: bool) -> None:
        """Configure the instrument and start or stop the acquisition operation.

        Parameters:
            reconfigure (bool): If True, the instrument settings are sent to the instrument.
            start (bool): If True, an acquisition is started. If False, an ongoing acquisition is stopped.

        Raises:
            DwfLibraryError: An error occurred while executing the *configure* operation.
        """
        result = self.lib.FDwfDigitalInConfigure(self.hdwf, reconfigure, start)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def status(self, read_data_flag: bool) -> DwfState:
        """Get the |DigitalIn| instrument state.

        This method performs a status request to the |DigitalIn| instrument and receives its response.

        The following methods can be used to retrieve |DigitalIn| instrument status information
        as a result of this call, regardless of the value of the *read_data_flag* parameter:

        * :py:meth:`~statusTime`
        * :py:meth:`~statusAutoTriggered`
        * :py:meth:`~statusSamplesLeft`
        * :py:meth:`~statusSamplesValid`
        * :py:meth:`~statusIndexWrite`
        * :py:meth:`~statusRecord`

        The following methods can be used to retrieve bulk data obtained from the |DigitalIn| instrument
        as a result of this call, but only if the *read_data_flag* parameter is True:

        * :py:meth:`~statusData`
        * :py:meth:`~statusData2`
        * :py:meth:`~statusNoise2`

        Parameters:
            read_data_flag (bool): Whether to read data.

        Returns:
            DwfState: The current state of the instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the *status* operation.
        """
        c_status = typespec_ctypes.DwfState()
        result = self.lib.FDwfDigitalInStatus(self.hdwf, read_data_flag, c_status)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        status = DwfState(c_status.value)
        return status

    ###################################################################################################################
    #                                                                                                                 #
    #                                                STATUS VARIABLES                                                 #
    #                                                                                                                 #
    ###################################################################################################################

    def statusTime(self) -> Tuple[int, int, int]:
        """Get the timestamp of the current status information.

        Returns:
            Tuple[int, int, int]: A three-element tuple, indicating the POSIX timestamp of the status request.
            The first element is the POSIX second, the second and third element are the numerator and denominator,
            respectively, of the fractional part of the second.

            In case :py:meth:`status` hasn't been called yet, this method will return zeroes
            for all three tuple elements.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_sec_utc = typespec_ctypes.c_unsigned_int()
        c_tick = typespec_ctypes.c_unsigned_int()
        c_ticks_per_second = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInStatusTime(
            self.hdwf,
            c_sec_utc,
            c_tick,
            c_ticks_per_second)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        sec_utc = c_sec_utc.value
        tick = c_tick.value
        ticks_per_second = c_ticks_per_second.value
        return (sec_utc, tick, ticks_per_second)

    def statusAutoTriggered(self) -> bool:
        """Check if the current acquisition is auto-triggered.

        Returns:
            bool: True if the current acquisition is auto-triggered, False otherwise.

        Raises:
            DwfLibraryError: An error occurred while retrieving the auto-triggered status.
        """
        c_auto_triggered = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusAutoTriggered(self.hdwf, c_auto_triggered)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        auto_triggered = bool(c_auto_triggered.value)
        return auto_triggered

    def statusSamplesLeft(self) -> int:
        """Retrieve the number of samples left in the acquisition, in samples.

        Returns:
            int: In case a finite-duration acquisition is active, the number of samples
            remaining to be acquired in the acquisition.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_samples_left = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusSamplesLeft(self.hdwf, c_samples_left)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        samples_left = c_samples_left.value
        return samples_left

    def statusSamplesValid(self) -> int:
        """Retrieve the number of valid data samples.

        Returns:
            int: The number of valid samples in the buffer.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_samples_valid = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusSamplesValid(self.hdwf, c_samples_valid)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        samples_valid = c_samples_valid.value
        return samples_valid

    def statusIndexWrite(self) -> int:
        """Retrieve the buffer write index.

        This is needed in :py:attr:`~pydwf.core.auxiliary.enum_types.DwfAcquisitionMode.ScanScreen`
        acquisition mode to display the scan bar.

        Returns:
            int: The buffer write index.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_index_write = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusIndexWrite(self.hdwf, c_index_write)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        index_write = c_index_write.value
        return index_write

    def statusRecord(self) -> Tuple[int, int, int]:
        """Get the recording status.

        Returns:
            Tuple[int, int, int]: A three-element tuple containing the counts for
            *available*, *lost*, and *corrupt* data samples, in that order.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_data_available = typespec_ctypes.c_int()
        c_data_lost = typespec_ctypes.c_int()
        c_data_corrupt = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusRecord(
            self.hdwf,
            c_data_available,
            c_data_lost,
            c_data_corrupt)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        data_free = c_data_available.value
        data_lost = c_data_lost.value
        data_corrupt = c_data_corrupt.value
        return (data_free, data_lost, data_corrupt)

    def statusCompress(self) -> Tuple[int, int, int]:
        """Get the recording status (compress version).

        Returns:
            Tuple[int, int, int]: A three-element tuple containing the counts for
            *available*, *lost*, and *corrupt* data samples, in that order.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_data_available = typespec_ctypes.c_int()
        c_data_lost = typespec_ctypes.c_int()
        c_data_corrupt = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInStatusCompress(
            self.hdwf,
            c_data_available,
            c_data_lost,
            c_data_corrupt)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        data_free = c_data_available.value
        data_lost = c_data_lost.value
        data_corrupt = c_data_corrupt.value
        return (data_free, data_lost, data_corrupt)

    ###################################################################################################################
    #                                                                                                                 #
    #                                              STATUS DATA RETRIEVAL                                              #
    #                                                                                                                 #
    ###################################################################################################################

    def statusData(self, count_bytes: int) -> np.ndarray:
        """Retrieve the acquired data samples from the |DigitalIn| instrument.

        Todo:
            Figure out the data format.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        samples = np.empty(count_bytes, dtype='B')
        result = self.lib.FDwfDigitalInStatusData(
            self.hdwf,
            samples.ctypes.data_as(typespec_ctypes.c_unsigned_char_ptr),
            count_bytes)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        return samples

    def statusData2(self, first_sample: int, count_bytes: int) -> np.ndarray:
        """Retrieve the acquired data samples from the |DigitalIn| instrument.

        Todo:
            Figure out the data format.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        samples = np.empty(count_bytes, dtype='B')
        result = self.lib.FDwfDigitalInStatusData2(
            self.hdwf,
            samples.ctypes.data_as(typespec_ctypes.c_unsigned_char_ptr),
            first_sample,
            count_bytes)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        return samples

    def statusCompressed(self, count_bytes: int) -> np.ndarray:
        """Retrieve the compressed data samples from the |DigitalIn| instrument.

        Todo:
            Figure out the data format.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        samples = np.empty(count_bytes, dtype='B')
        result = self.lib.FDwfDigitalInStatusData(
            self.hdwf,
            samples.ctypes.data_as(typespec_ctypes.c_unsigned_char_ptr),
            count_bytes)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        return samples


    def statusCompressed2(self, first_sample: int, count_bytes: int) -> np.ndarray:
        """Retrieve the acquired data samples from the |DigitalIn| instrument.

        Todo:
            Figure out the data format.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        samples = np.empty(count_bytes, dtype='B')
        result = self.lib.FDwfDigitalInStatusCompressed2(
            self.hdwf,
            samples.ctypes.data_as(typespec_ctypes.c_unsigned_char_ptr),
            first_sample,
            count_bytes)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        return samples

    def statusNoise2(self, first_sample: int, count_bytes: int) -> np.ndarray:
        """Get the noise data from the |DigitalIn| instrument.

        Todo:
            Figure out the data format.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        noise = np.empty(count_bytes, dtype='B')
        result = self.lib.FDwfDigitalInStatusNoise2(
            self.hdwf,
            noise.ctypes.data_as(typespec_ctypes.c_unsigned_char_ptr),
            first_sample,
            count_bytes)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        return noise

    ###################################################################################################################
    #                                                                                                                 #
    #                                           ACQUISITION TIMING SETTINGS                                           #
    #                                                                                                                 #
    ###################################################################################################################

    def internalClockInfo(self) -> float:
        """Get the |DigitalIn| internal clock frequency, in Hz.

        Returns:
            The internal clock frequency, in Hz.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_frequency = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalInInternalClockInfo(self.hdwf, c_frequency)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        frequency = c_frequency.value
        return frequency

    def clockSourceInfo(self) -> List[DwfDigitalInClockSource]:
        """Get a list of valid clock sources for the |DigitalIn| instrument.

        Returns:
            List[DwfDigitalInClockSource]: A list of valid clock sources.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_clock_source_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInClockSourceInfo(self.hdwf, c_clock_source_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        clock_source_bitset = c_clock_source_bitset.value
        clock_source_list = [clock_source for clock_source in DwfDigitalInClockSource
                             if clock_source_bitset & (1 << clock_source.value)]
        return clock_source_list

    def clockSourceSet(self, clock_source: DwfDigitalInClockSource) -> None:
        """Set the |DigitalIn| instrument clock source.

        Parameters:
            clock_source (DwfDigitalInClockSource): The clock source to be selected.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInClockSourceSet(self.hdwf, clock_source.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def clockSourceGet(self) -> DwfDigitalInClockSource:
        """Get the |DigitalIn| instrument clock source.

        Returns:
            DwfDigitalInClockSource: The currently configured |DigitalIn| clock source.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_clock_source = typespec_ctypes.DwfDigitalInClockSource()
        result = self.lib.FDwfDigitalInClockSourceGet(self.hdwf, c_clock_source)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        clock_source = DwfDigitalInClockSource(c_clock_source.value)
        return clock_source

    def dividerInfo(self) -> int:
        """Get the |DigitalIn| instrument maximum divider value.

        Returns:
            int: The maximum valid divider value that can be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_divider_max = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInDividerInfo(self.hdwf, c_divider_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        divider_max = c_divider_max.value
        return divider_max

    def dividerSet(self, divider: int) -> None:
        """Set the |DigitalIn| instrument divider value.

        Parameters:
            divider: The divider value to be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInDividerSet(self.hdwf, divider)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def dividerGet(self) -> int:
        """Get the current |DigitalIn| instrument divider value.

        If the clock source is internal, the |DigitalIn| sample frequency will be equal
        to :py:meth:`internalClockInfo` divided by :py:meth:`dividerGet`.

        Returns:
            int: The currently configured divider value.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_divider = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInDividerGet(self.hdwf, c_divider)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        divider = c_divider.value
        return divider

    ###################################################################################################################
    #                                                                                                                 #
    #                                              ACQUISITION SETTINGS                                               #
    #                                                                                                                 #
    ###################################################################################################################

    def acquisitionModeInfo(self) -> List[DwfAcquisitionMode]:
        """Get a list of valid |DigitalIn| instrument acquisition modes.

        Returns:
            List[DwfAcquisitionMode]: A list of valid acquisition modes for the |DigitalIn| instrument.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_acquisition_mode_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInAcquisitionModeInfo(self.hdwf, c_acquisition_mode_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        acquisition_mode_bitset = c_acquisition_mode_bitset.value
        acquisition_mode_list = [acquisition_mode for acquisition_mode in DwfAcquisitionMode
                                 if acquisition_mode_bitset & (1 << acquisition_mode.value)]
        return acquisition_mode_list

    def acquisitionModeSet(self, acquisition_mode: DwfAcquisitionMode) -> None:
        """Select the |DigitalIn| acquisition mode.

        Parameters:
            acquisition_mode (DwfAcquisitionMode): The acquisition mode to be selected.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInAcquisitionModeSet(self.hdwf, acquisition_mode.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def acquisitionModeGet(self) -> DwfAcquisitionMode:
        """Get the currently selected |DigitalIn| acquisition mode.

        Returns:
            DwfAcquisitionMode: The currently selected acquisition mode.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_acquisition_mode = typespec_ctypes.DwfAcquisitionMode()
        result = self.lib.FDwfDigitalInAcquisitionModeGet(self.hdwf, c_acquisition_mode)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        acquisition_mode = DwfAcquisitionMode(c_acquisition_mode.value)
        return acquisition_mode

    def bitsInfo(self) -> int:
        """Get the number of |DigitalIn| bits.

        Returns:
            The number of digital input bits available.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_num_bits = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInBitsInfo(self.hdwf, c_num_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        num_bits = c_num_bits.value
        return num_bits

    def sampleFormatSet(self, num_bits: int) -> None:
        """Set the |DigitalIn| sample format (i.e., number of bits).

        Parameters:
            num_bits (int): The number of bits per sample (8, 16, or 32).

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInSampleFormatSet(self.hdwf, num_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def sampleFormatGet(self) -> int:
        """Get the |DigitalIn| sample format (i.e., number of bits).

        Returns:
            int: The currently configured number of bits per sample (8, 16, or 32).

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_num_bits = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInSampleFormatGet(self.hdwf, c_num_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        num_bits = c_num_bits.value
        return num_bits

    def inputOrderSet(self, dio_first: bool) -> None:
        """Select the |DigitalIn| order of values stored in the sampling array.

        If *dio_first* is True, DIO 24…39 are placed at the beginning of the array followed by DIN 0…23.

        If *dio_first* is False, DIN 0…23 are placed at the beginning followed by DIO 24…31.

        This method is valid only for the Digital Discovery device.

        Parameters:
            dio_first (bool): Whether the DIO pins come before the DIN pins.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInInputOrderSet(self.hdwf, int(dio_first))
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def bufferSizeInfo(self) -> int:
        """Get the |DigitalIn| instrument maximum buffer size.

        Returns:
            int: The maximum valid buffer size.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_buffer_size_max = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInBufferSizeInfo(self.hdwf, c_buffer_size_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        buffer_size_max = c_buffer_size_max.value
        return buffer_size_max

    def bufferSizeSet(self, buffer_size: int) -> None:
        """Set the |DigitalIn| instrument buffer size.

        Parameters:
            buffer_size (int): The buffer size to be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInBufferSizeSet(self.hdwf, buffer_size)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def bufferSizeGet(self) -> int:
        """Get the currently configured |DigitalIn| instrument buffer size.

        Returns:
            int: The currently configured buffer size.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_buffer_size = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInBufferSizeGet(self.hdwf, c_buffer_size)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        buffer_size = c_buffer_size.value
        return buffer_size

    def sampleModeInfo(self) -> List[DwfDigitalInSampleMode]:
        """Get the valid |DigitalIn| instrument sample modes.

        Returns:
            List[DwfDigitalInSampleMode]: A list of valid sample modes.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_sample_mode_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInSampleModeInfo(self.hdwf, c_sample_mode_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        sample_mode_bitset = c_sample_mode_bitset.value
        sample_mode_list = [sample_mode for sample_mode in DwfDigitalInSampleMode
                            if sample_mode_bitset & (1 << sample_mode.value)]
        return sample_mode_list

    def sampleModeSet(self, sample_mode: DwfDigitalInSampleMode) -> None:
        """Set the |DigitalIn| instrument sample mode.

        Parameters:
            sample_mode (DwfDigitalInSampleMode): The sample mode to be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInSampleModeSet(self.hdwf, sample_mode.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def sampleModeGet(self) -> DwfDigitalInSampleMode:
        """Get the |DigitalIn|  instrument sample mode.

        Returns:
            DwfDigitalInSampleMode: The currently configured sample mode.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_sample_mode = typespec_ctypes.DwfDigitalInSampleMode()
        result = self.lib.FDwfDigitalInSampleModeGet(self.hdwf, c_sample_mode)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        sample_mode = DwfDigitalInSampleMode(c_sample_mode.value)
        return sample_mode

    def sampleSensibleSet(self, compression_bits: int) -> None:
        """Set the |DigitalIn| instrument *sample sensible* setting.

        Todo:
            Figure out what this setting does.

        Parameters:
            compression_bits (int): (unknown)

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInSampleSensibleSet(self.hdwf, compression_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def sampleSensibleGet(self) -> int:
        """Get the |DigitalIn| instrument *sample sensible* setting.

        This setting is only used in :py:attr:`~pydwf.core.auxiliary.enum_types.DwfAcquisitionMode.Record` mode.

        It select the signals to be used for data compression in record acquisition mode.

        Todo:
            Figure out what this setting does.

        Returns:
            int: The sample sensible setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_compression_bits = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInSampleSensibleGet(self.hdwf, c_compression_bits)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        compression_bits = c_compression_bits.value
        return compression_bits

    def triggerPrefillSet(self, samples_before_trigger: int) -> None:
        """Set the |DigitalIn| instrument trigger prefill setting, in samples.

        This setting is only used in :py:attr:`~pydwf.core.auxiliary.enum_types.DwfAcquisitionMode.Record` mode.

        It determines the number of samples to acquire before arming in
        :py:attr:`~pydwf.core.auxiliary.enum_types.DwfAcquisitionMode.Record` acquisition mode. The prefill is used for
        recording with a trigger to make sure that at least the required number of samples are collected before arming.

        Todo:
            Figure out what this setting does, precisely.

        Parameters:
            samples_before_trigger (int): The prefill count, in samples.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerPrefillSet(self.hdwf, samples_before_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerPrefillGet(self) -> int:
        """Get the |DigitalIn| instrument trigger prefill setting, in samples.

        Returns:
            int: The trigger prefill count, in samples.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_samples_before_trigger = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInTriggerPrefillGet(self.hdwf, c_samples_before_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        samples_before_trigger = c_samples_before_trigger.value
        return samples_before_trigger

    ###################################################################################################################
    #                                                                                                                 #
    #                                              TRIGGER CONFIGURATION                                              #
    #                                                                                                                 #
    ###################################################################################################################

    def triggerSourceInfo(self) -> List[DwfTriggerSource]:
        """Get the valid |DigitalIn| instrument trigger sources.

        Warning:
            **This method is obsolete.**

            Use the generic :py:meth:`DeviceControl.triggerInfo()` method instead.

        Returns:
            List[DwfTriggerSource]: A list of valid trigger sources.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_source_bitset = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInTriggerSourceInfo(self.hdwf, c_trigger_source_bitset)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_source_bitset = c_trigger_source_bitset.value
        trigger_source_list = [trigger_source for trigger_source in DwfTriggerSource
                               if trigger_source_bitset & (1 << trigger_source.value)]
        return trigger_source_list

    def triggerSourceSet(self, trigger_source: DwfTriggerSource) -> None:
        """Set |DigitalIn| instrument trigger source.

        Parameters:
            trigger_source (DwfTriggerSource): The trigger source to be configured.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerSourceSet(self.hdwf, trigger_source.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerSourceGet(self) -> DwfTriggerSource:
        """Get the currently selected instrument trigger source.

        Returns:
            DwfTriggerSource: The currently selected instrument trigger source.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_source = typespec_ctypes.DwfTriggerSource()
        result = self.lib.FDwfDigitalInTriggerSourceGet(self.hdwf, c_trigger_source)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_source = DwfTriggerSource(c_trigger_source.value)
        return trigger_source

    def triggerSlopeSet(self, trigger_slope: DwfTriggerSlope) -> None:
        """Select the |DigitalIn| instrument trigger slope.

        Parameters:
            trigger_slope (DwfTriggerSlope): The trigger slope to be selected.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerSlopeSet(self.hdwf, trigger_slope.value)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerSlopeGet(self) -> DwfTriggerSlope:
        """Get the currently selected |DigitalIn| instrument trigger slope.

        Returns:
            DwfTriggerSlope: The currently selected |DigitalIn| instrument trigger slope.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_slope = typespec_ctypes.DwfTriggerSlope()
        result = self.lib.FDwfDigitalInTriggerSlopeGet(self.hdwf, c_trigger_slope)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        trigger_slope = DwfTriggerSlope(c_trigger_slope.value)
        return trigger_slope

    def triggerPositionInfo(self) -> int:
        """Get |DigitalIn| trigger position info.

        Returns:
            int: The maximum number of samples after the trigger.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_max_samples_after_trigger = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInTriggerPositionInfo(
            self.hdwf,
            c_max_samples_after_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        max_samples_after_trigger = c_max_samples_after_trigger.value
        return max_samples_after_trigger

    def triggerPositionSet(self, samples_after_trigger: int) -> None:
        """Set |DigitalIn| instrument desired trigger position.

        Parameters:
            samples_after_trigger (int): The number of samples after the trigger.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerPositionSet(self.hdwf, samples_after_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerPositionGet(self) -> int:
        """Get |DigitalIn| instrument trigger position.

        Returns:
            int: The currently configured number of samples after the trigger.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_samples_after_trigger = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInTriggerPositionGet(self.hdwf, c_samples_after_trigger)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        samples_after_trigger = c_samples_after_trigger.value
        return samples_after_trigger

    def triggerAutoTimeoutInfo(self) -> Tuple[float, float, int]:
        """Get |DigitalIn| instrument trigger auto-timeout range, in seconds.

        Returns:
            Tuple[float, float, int]: The range and number of steps of the auto-timeout setting.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_trigger_auto_timeout_min = typespec_ctypes.c_double()
        c_trigger_auto_timeout_max = typespec_ctypes.c_double()
        c_trigger_auto_timeout_num_steps = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalInTriggerAutoTimeoutInfo(
            self.hdwf,
            c_trigger_auto_timeout_min,
            c_trigger_auto_timeout_max,
            c_trigger_auto_timeout_num_steps)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        if not c_trigger_auto_timeout_num_steps.value.is_integer():
            raise PyDwfError("Bad c_trigger_auto_timeout_num_steps value.")

        trigger_auto_timeout_min = c_trigger_auto_timeout_min.value
        trigger_auto_timeout_max = c_trigger_auto_timeout_max.value
        trigger_auto_timeout_num_steps = int(c_trigger_auto_timeout_num_steps.value)
        return (trigger_auto_timeout_min, trigger_auto_timeout_max, trigger_auto_timeout_num_steps)

    def triggerAutoTimeoutSet(self, auto_timeout: float) -> None:
        """Set |DigitalIn| instrument trigger auto-timeout value, in seconds.

        Parameters:
            auto_timeout (float): The auto-timeout value to be configured, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerAutoTimeoutSet(self.hdwf, auto_timeout)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerAutoTimeoutGet(self) -> float:
        """Get |DigitalIn| instrument trigger auto-timeout value, in seconds.

        Returns:
            float: The currently configured trigger auto-timeout value, in seconds.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_auto_timeout = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalInTriggerAutoTimeoutGet(self.hdwf, c_auto_timeout)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        auto_timeout = c_auto_timeout.value
        return auto_timeout

    ###################################################################################################################
    #                                                                                                                 #
    #                                         TRIGGER DETECTOR CONFIGURATION                                          #
    #                                                                                                                 #
    ###################################################################################################################

    def triggerInfo(self) -> Tuple[int, int, int, int]:
        """Get |DigitalIn| detector trigger info.

        Return the pins that support the different types of triggering. Each integer is a bitmask, representing pins
        that can be used as *low level*, *high level*, *rising edge*, and *falling edge* triggers.

        Returns:
            Tuple[int, int, int, int]: Digital pins that can be used as *low level*, *high level*, *rising edge*,
            and *falling edge* triggers, respectively.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_level_low = typespec_ctypes.c_unsigned_int()
        c_level_high = typespec_ctypes.c_unsigned_int()
        c_edge_rise = typespec_ctypes.c_unsigned_int()
        c_edge_fall = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInTriggerInfo(
            self.hdwf,
            c_level_low,
            c_level_high,
            c_edge_rise,
            c_edge_fall)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        level_low = c_level_low.value
        level_high = c_level_high.value
        edge_rise = c_edge_rise.value
        edge_fall = c_edge_fall.value
        return (level_low, level_high, edge_rise, edge_fall)

    def triggerSet(self, level_low: int, level_high: int, edge_rise: int, edge_fall: int) -> None:
        """Set |DigitalIn| trigger detector conditions.

        The *level_low* and *level_high* settings effectively mask trigger detection events to clock cycles where
        the selected pins are low or high, respectively.

        The *edge_rise* and *edge_fall* indicate channels where the specific type of transition on any of the
        included channels will lead to a trigger event, ar least if the level conditions specified by the *level_low*
        and *level_high* settings are satisfied.

        Parameters:
            level_low (int): The channels that are required to be low for a trigger event to occur (bitfield).
            level_high (int): The channels that are required to be high for a trigger event to occur (bitfield).
            edge_rise (int): The channels where a rising edge will cause a trigger event (bitfield).
            edge_fall (int): The channels where a falling edge will cause a trigger event (bitfield).

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerSet(
            self.hdwf,
            level_low,
            level_high,
            edge_rise,
            edge_fall)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerGet(self) -> Tuple[int, int, int, int]:
        """Get the configured |DigitalIn| trigger detector settings.

        Returns:
            Tuple[int, int, int, int]: The currently configured *level_low*, *level_high*, *edge_rise*,
            and *edge_fall* settings.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        c_level_low  = typespec_ctypes.c_unsigned_int()
        c_level_high = typespec_ctypes.c_unsigned_int()
        c_edge_rise  = typespec_ctypes.c_unsigned_int()
        c_edge_fall  = typespec_ctypes.c_unsigned_int()
        result = self.lib.FDwfDigitalInTriggerGet(
            self.hdwf,
            c_level_low,
            c_level_high,
            c_edge_rise,
            c_edge_fall)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        level_low  = c_level_low.value
        level_high = c_level_high.value
        edge_rise  = c_edge_rise.value
        edge_fall  = c_edge_fall.value
        return (level_low, level_high, edge_rise, edge_fall)

    def triggerResetSet(self, level_low: int, level_high: int, edge_rise: int, edge_fall: int) -> None:
        """Configure |DigitalIn| trigger detector reset condition.

        Parameters:
            level_low (int): The channels that are required to be low for a reset to occur (bitfield).
            level_high (int): The channels that are required to be high for a reset to occur (bitfield).
            edge_rise (int): The channels where a rising edge will cause a reset (bitfield).
            edge_fall (int): The channels where a falling edge will cause a reset (bitfield).

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerResetSet(
            self.hdwf,
            level_low,
            level_high,
            edge_rise,
            edge_fall)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerCountSet(self, count: int, restart: int) -> None:
        """Set the |DigitalIn| trigger detector count.

        Todo:
            Figure out what this setting does.

        Parameters:
            count (int): The event count.
            restart (int): (to be documented)

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerCountSet(self.hdwf, count, restart)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerLengthSet(self, min_length: float, max_length: float, sync_mode: int) -> None:
        """Set the |DigitalIn| trigger detector length.

        Todo:
            Figure out what this setting does.

        Parameters:
            min_length (float): (to be documented)
            max_length (float): (to be documented)
            sync_mode (int): Synchronization mode:

                * 0 — Normal.
                * 1 — Timing: the *min_length* parameter specifies the bit length and the *max_length* parameter
                  specifies the timing length.

                  Used for UART, CAN protocols.
                * 2 — PWM: the *min_length* parameter specifies the sampling time and the *max_length* parameter
                  specifies the timing length.

                  Used for 1-wire protocols.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerLengthSet(self.hdwf, min_length, max_length, sync_mode)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def triggerMatchSet(self, pin: int, mask: int, value: int, bit_stuffing: int) -> None:
        """Set |DigitalIn| trigger detector match.

        Todo:
            Figure out what this setting does.

        Configure the deserializer. The bits are left shifted.
        The mask and value should be specified accordingly, in MSB-first order.

        Parameters:
            pin (int): The pin to be deserialized.
            mask (int): The mask pattern.
            value (int): The bit pattern.
            bit_stuffing (int): The bit stuffing count.

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInTriggerMatchSet(self.hdwf, pin, mask, value, bit_stuffing)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    ###################################################################################################################
    #                                                                                                                 #
    #                                                     COUNTER                                                     #
    #                                                                                                                 #
    ###################################################################################################################

    def counterInfo(self) -> Tuple[int, float]:
        """Get |DigitalIn| counter info."""
        c_counter_max = typespec_ctypes.c_double()
        c_duration_max = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalInCounterInfo(self.hdwf, c_counter_max, c_duration_max)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

        if not c_counter_max.value.is_integer():
            raise PyDwfError("Bad c_counter_max value.")

        counter_max = int(c_counter_max.value)
        duration_max = c_duration_max.value
        return (counter_max, duration_max)

    def counterSet(self, duration: float) -> None:
        """Set |DigitalIn| counter duration."""
        result = self.lib.FDwfDigitalInCounterSet(self.hdwf, duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()

    def counterGet(self) -> float:
        """Get |DigitalIn| counter duration."""
        c_duration = typespec_ctypes.c_double()
        result = self.lib.FDwfDigitalInCounterGet(self.hdwf, c_duration)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        duration = c_duration.value
        return duration

    def counterStatus(self) -> Tuple[float, float, int]:
        """Get |DigitalIn| counter status."""
        c_count = typespec_ctypes.c_double()
        c_freq = typespec_ctypes.c_double()
        c_tick = typespec_ctypes.c_int()
        result = self.lib.FDwfDigitalInCounterStatus(self.hdwf, c_count, c_freq, c_tick)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()
        count = c_count.value
        freq = c_freq.value
        tick = c_tick.value
        return (count, freq, tick)

    ###################################################################################################################
    #                                                                                                                 #
    #                                             MISCELLANEOUS SETTINGS                                              #
    #                                                                                                                 #
    ###################################################################################################################

    def mixedSet(self, enable: bool) -> None:
        """Set mixed state.

        Warning:
            **This method is obsolete.**

        Todo:
            Figure out what this setting does.

        Parameters:
            enable (bool): (to be documented)

        Raises:
            DwfLibraryError: An error occurred while executing the operation.
        """
        result = self.lib.FDwfDigitalInMixedSet(self.hdwf, enable)
        if result != RESULT_SUCCESS:
            raise self.dwf.exception()



"""The |pydwf.core.dwf_device_protocol_apis| module provides a single class: |DwfDeviceProtocolAPIs|."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pydwf.core.api.protocol_uart import ProtocolUART
from pydwf.core.api.protocol_spi import ProtocolSPI
from pydwf.core.api.protocol_i2c import ProtocolI2C
from pydwf.core.api.protocol_can import ProtocolCAN
from pydwf.core.api.protocol_swd import ProtocolSWD

if TYPE_CHECKING:
    # This is only needed when type-checking.
    from pydwf.core.dwf_device import DwfDevice


class DwfDeviceProtocolAPIs:
    """The |DwfDeviceProtocolAPIs| wraps the five digital protocols.

    Attention:

        Users of |pydwf| should not create instances of this class directly.

        Use |DeviceControl.open:link| or |pydwf.utilities.openDwfDevice:link| to obtain a valid
        |DwfDevice| instance.

    The main test and measurement functionality of a Digilent Waveforms device is provided as
    multiple sub-interfaces (instruments, protocols, and measurements). To access those,
    use one of the twelve attributes described below.

    .. rubric:: |DwfDevice| attributes

    Attributes:

        uart (ProtocollUART):
            Provides access to the UART protocol functionality.
        can (DigitalCAN):
            Provides access to the CAN protocol functionality.
        spi (DigitalSPI):
            Provides access to the SPI protocol functionality.
        i2c (DigitalI2C):
            Provides access to the I²C protocol functionality.
        swd (DigitalSWD):
            Provides access to the SWD protocol functionality.

    .. rubric:: |DwfDevice| properties and methods
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, device: DwfDevice) -> None:
        """Initialize a DwfDeviceProtocols with the specified DWF handle value.

        This method is used by the |DwfLibrary.device.open| method to initialize a |DwfDevice| immediately after
        successfully opening a Digilent Waveforms device using the low-level C API.

        User programs should not initialize |DwfDeviceProtocols| instances themselves,
        but leave that to |pydwf| instead.

        Parameters:
            dwf (DwfLibrary):
                The |DwfLibrary| instance used to make calls to the underlying library.
            hdwf (int):
                The device to open.
        """

        self.uart = ProtocolUART(device)
        self.spi = ProtocolSPI(device)
        self.i2c = ProtocolI2C(device)
        self.can = ProtocolCAN(device)
        self.swd = ProtocolSWD(device)
